/***************************************************************************
                          r2a_file.cpp  -  description
                             -------------------
    begin                : Sun Jun 6 2004
    copyright            : (C) 2004 by Harm Anne Schotanus
    email                : spearhead@wanadoo.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <string>
#include <fstream>
#include <cctype>
#include <sstream>
#include <iterator>
#include "riv2ascii.h"
#include "riv_error.h"
#include "r2a_country.h"
#include "r2a_sub.h"
#include "r2a_helper.h"
#include "r2a_main.h"
#include "r2a_country_list.h"
#include "r2a_sub_list.h"
#include "r2a_main_list.h"
#include "r2a_regime.h"
#include "r2a_regime_list.h"
#include "r2a_uicnr.h"
#include "r2a_speed_heat_list.h"
#include "r2a_debug.h"
#include "r2a_main_type_list.h"
#include "r2a_speed_heat.h"
#include "r2a_main_type.h"
#include "r2a_options.h"
#include "r2a_files.h"
#include "r2a_iga.h"
#include "r2a_iga_list.h"
#include "r2a_file_error.h"


extern regime_list regimes;
extern main_class_list main_classes;
extern country_code_list country_codes;
extern speed_heat_list speeds_heats;
extern main_type_list main_types;
extern iga_list igas;


using std::ifstream;
using std::cerr;


namespace {


void parse_igad (
  const string& s,
  const unsigned line,
  const optionsT& options
)
{
  print_debug (cerr, 3, options, "parsing line: ", s);
  try {
    handleT <iga> i= handleT <iga> (new iga());
    i->parse_iga(s);
    igas.add(i);
  }
  catch (parse_error) {
    if (line > 1) throw;
  }
}


void
parse_bounds (
  const string& s,
  unsigned& min_code,
  unsigned& max_code,
  const string::size_type pos=0,
  const string::size_type len=string::npos
)
{
  if (s[pos] == '-') {
    min_code = 0;
    max_code = convert_num (s, pos + 1);
    return ;
  }
  bool es;
  min_code = convert_num (s, es, pos, len);
  if (es) {
    max_code = min_code;
    return;
  }
  string::size_type tidx = s.find_first_of ("+-", pos);
  if (tidx == string::npos) throw parse_error(s);
  switch (s[tidx]) {
  case '-':
      max_code = convert_num (s, tidx + 1);
      break;
  case '+':
      max_code = 0;
      break;
  default:
      throw parse_error(s);
      break;
  }
}


string::size_type
find(
  const string& s,
  const char c,
  const string::size_type bidx =0
)
{
  string::size_type eidx = s.find(c, bidx);
  if (eidx == string::npos) throw parse_error(s);
  return eidx;
}


void
parse_regime (
  const string& s,
  const optionsT& options
)
/*
# format : <nr>-<nr>,F,<pool>,<axles>-<axles>|<axles>+,<owner>
# or     : <nr>-<nr>,P,<pool>,(<extra>)
# or 	 : <nr>-<nr>,L,(extra)
*/
{
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find(s, ',');

  unsigned max_code, min_code;
  parse_bounds(s, min_code, max_code, 0, eidx);
  string::size_type bidx = eidx + 1;
  eidx = find(s, ',', bidx);
  switch (toupper (s[bidx])) {
  case 'F':{
    bidx = eidx + 1;
    eidx = find(s, ',', bidx);
    bidx = skip_spaces(s, bidx, eidx - bidx);
    string pool = s.substr(bidx, eidx - bidx);
    bidx = eidx + 1;
    eidx = find(s, ',', bidx);
    unsigned min_axles, max_axles;
    parse_bounds(s, min_axles, max_axles, bidx, eidx - bidx);
    bidx = skip_spaces(s, eidx + 1);
    string owner = s.substr(bidx);
    regimes.add_goods(min_code, max_code, min_axles, max_axles, pool, owner);
    break;
    }
  case 'L': {
    string extra = s.substr(eidx + 1);
    regimes.add_loc (min_code, max_code, extra);
    break;
    }
  case 'P': {
    bidx = eidx + 1;
    eidx = find(s, ',', bidx);
    bidx = skip_spaces(s, bidx, eidx - bidx);
    string pool = s.substr(bidx, eidx - bidx);
    string extra;
    if (eidx != string::npos) {
      bidx = skip_spaces(s, eidx + 1);
      extra = s.substr(bidx);
    }
    regimes.add_passenger(min_code, max_code, pool, extra);
    break;
    }
  default:
    throw parse_error(s);
    break;
  }
}


void
parse_speed_heat(
  const string& s,
  const optionsT& options
)
{
/*
  #format <nr>-<nr>,speed, heating
*/
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find(s, ',');
  unsigned max_code, min_code;
  parse_bounds(s, min_code, max_code, 0, eidx);
  string::size_type bidx = eidx + 1;
  eidx = find(s, ',', bidx);
  unsigned speed = convert_num(s, bidx);
  string heating = s.substr(eidx + 1);
  speeds_heats.add(handleT <speed_heat> (new speed_heat (min_code, max_code, speed, heating)));
}


void
parse_main_type(
  const string& s,
  const optionsT& options
)
{
/*
#format : <nr>-<nr>,F,<capital>
# or    : <nr>-<nr>,P,<descrition>
*/
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find(s, ',');
  unsigned max_code, min_code;
  parse_bounds(s, min_code, max_code, 0, eidx);
  string::size_type bidx = eidx + 1;
  eidx = find(s, ',', bidx);
  bidx = skip_spaces (s, bidx, eidx);
  eidx = skip_spaces (s, eidx + 1);
  main_type::car_typeT type;
  string description;
  switch (toupper (s[bidx])) {
  case 'F':{
    type = main_type::Freight;
    char c = s[eidx];
    if (!isupper(c)) throw inv_letter (c);
    description.push_back(c);
    break;
    }
  case 'P':
    type = main_type::Passenger;
    description = s.substr(eidx);
    break;
  default:
    throw parse_error(s);
  }
  main_types.add(handleT <main_type> (new main_type (min_code, max_code, type, description)));
}


void
parse_country (
  const string& s,
  const optionsT& options
)
{
   	/* format: Country code, abbrev, first used, last used, country, company */
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find (s, ',');
	unsigned cnt_code = convert_num (s);
  string::size_type bidx = eidx + 1;
	eidx = find (s, ',', bidx);
  bidx = s.find_first_not_of ("\t ", bidx);
  if (bidx == string::npos) throw parse_error(s);
  bidx = skip_spaces (s, bidx, eidx);
  string abbr = s.substr (bidx, eidx - bidx);
  bidx = eidx + 1;
	eidx = find (s,',', bidx);
  unsigned max_year, min_year;
  parse_bounds(s, min_year, max_year, bidx, eidx - bidx);
  bidx = eidx + 1;
  eidx = find (s, ',', bidx);
  bidx = skip_spaces (s, bidx, eidx);
  string cntry = s.substr (bidx, eidx - bidx);
  eidx = skip_spaces (s, eidx + 1);
  string company = s.substr (eidx);
	country_codes.add(handleT<country_code> (new country_code (cnt_code, cntry, abbr, company, min_year, max_year)));
}


void
parse_capital(
  const string& s,
  const optionsT& options
)
{
/*format: Capital letter,wheel sets (lowerbound),wheel sets (upperbound),year( low),
	year( up), Description
 	0 means no bound
e.g. E,2,2,1975,0;some car
or
     B,0,0,0,0; another car
*/
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find (s, ',');
  string::size_type bidx = skip_spaces (s, 0, eidx);
  if (!isupper(s[bidx])) throw parse_error(s);
  char l = s[bidx];
  bidx = eidx + 1;
  eidx = find (s, ',', bidx);
  unsigned min_axles,max_axles;
  parse_bounds (s, min_axles, max_axles, bidx, eidx - bidx);
  bidx = eidx + 1 ;
  eidx = find (s, ';', bidx);
  unsigned min_year, max_year;
  parse_bounds (s, min_year, max_year, bidx, eidx - bidx);
  eidx = skip_spaces (s, eidx + 1);
  string descr = s.substr(eidx ) ;
  main_classes.add (handleT <main_class> (new main_class(l, descr, min_axles, max_axles, min_year, max_year)));
}


void
parse_small (
  const string& s,
  const unsigned line,
  const optionsT& options
)
{
/* format: letter (or two), matching capitals, UIC country code, year-year, wheelsets-wheelsets,car-type, description
0 for country code means match all
0 for year means no limit
* for capitals means match all

last element before ; : a - articulated, s - single , u - wagon unit, other character indifferent


Eg: a,AES,84,1975-2000,2-6,a;special car
or
    ss,KS,0,0,0,b;some other special car;	

*/
  print_debug (cerr, 3, options, "parsing line: ", s);
  string::size_type eidx = find (s, ',');
  string::size_type bidx = skip_spaces (s, 0, eidx);
  if (!islower (s[bidx])) throw parse_error(s, line);

  char letter = s[bidx];
  unsigned doubled = 0;
  for (; (s[bidx + 1]) == letter; ++bidx) ++doubled;
  bidx = eidx + 1;
  eidx = find (s, ',', bidx);
  string caps;

 	for (string::size_type tmp = bidx; tmp <= eidx; ++tmp) {
		if (s[tmp] == '*') {
			caps = "*";
		  break;
		} else if (s[tmp] == '(') {
      ++tmp;
      if (!isupper(s[tmp])) throw parse_error(s, line);
      while (tmp <= eidx && s[tmp] !=	')') ++tmp;
      if (s[tmp] != ')') throw parse_error(s, line);
    } else if (!isupper(s[tmp])) {
      caps = s.substr(bidx, tmp - bidx);
      break;
    }
  }

  bidx = eidx + 1;
  eidx = find (s, ',', bidx);
  unsigned cntry = convert_num(s, bidx);
  bidx = eidx + 1;
  eidx = find (s,',', bidx);
  unsigned min_year, max_year ;
  parse_bounds(s, min_year, max_year, bidx, eidx - bidx);
  bidx = eidx + 1;
  eidx = find (s, ',', bidx);
  unsigned max_axles, min_axles;
  parse_bounds(s, min_axles, max_axles, bidx, eidx - bidx);
  bidx = eidx + 1;
  eidx = find (s, ';', bidx);
  typeT tp = maketype (s[bidx]);
  eidx = skip_spaces(s, eidx + 1);
  string descr = s.substr (eidx);
	main_classes.add_sub(caps, letter, doubled, cntry, min_year, max_year, min_axles, max_axles, tp, descr, line);
}


namespace file_strings {


const string CAPITALSTR("capitals");
const string SMALLSTR("small");
const string COUNTRYSTR("country");
const string REGIMESTR("regime");
const string SPEEDHEATSTR("speed&heat");
const string MAINTYPESTR("maintype");
const string IGASTR("iga");
const string NONESTR("none");


} //namespace


filetypeT
get_type_str(
  const string& s,
  const unsigned offset
)
{
  if (string_compare (s,file_strings::CAPITALSTR, offset, 8) == 0) return UIC_CAPITAL;
  else if (string_compare (s,file_strings::SMALLSTR, offset, 5) == 0) return UIC_SMALL;
  else if (string_compare (s,file_strings::COUNTRYSTR, offset, 7) == 0) return UIC_COUNTRY;
  else if (string_compare (s,file_strings::REGIMESTR, offset, 6) == 0) return UIC_REGIME;
  else if (string_compare (s,file_strings::SPEEDHEATSTR, offset, 10) == 0) return UIC_SPEEDHEAT;
  else if (string_compare (s,file_strings::MAINTYPESTR, offset, 8) == 0) return UIC_MAINTYPE;
  else if (string_compare (s,file_strings::IGASTR, offset, 3) == 0) return UIC_IGA;
  else if (string_compare (s,file_strings::NONESTR, offset, 4) == 0) return UIC_UNDEFINED;
  else throw ill_value(ill_value::TYPE);
}


} // namespace


namespace file_cmd_strings {


const string AXLESSTR ("axles:");
const string COUNTRYSTR ("country:");
const string HTMLSTR ("html:");
const string INCLUDESTR  ("include:");
const string FILETYPESTR ("filetype:");
const string TYPESTR ("type:");
const string YEARSTR ("year:");
const string DOIGASTR ("doiga:");
const string DEBUGSTR ("debug:");
const string HISTORYSTR ("history:");


} // namespace


bool
read_commands(
  const string &s,
  optionsT& options,
  filetypeT& tp
)
{
  if (string_compare (s, file_cmd_strings::FILETYPESTR, 0, 9) == 0)
    tp = get_type_str(s, 9);
  else if (string_compare (s, file_cmd_strings::INCLUDESTR, 0, 8) == 0)
    read_file (s.c_str() + 8, options, tp);
  else if (string_compare (s, file_cmd_strings::HTMLSTR, 0, 5) == 0)
    options.html = parse_bool(s, 5);
  else if (string_compare (s, file_cmd_strings::DOIGASTR, 0, 6) == 0)
    options.show_iga = parse_bool(s, 6);
  else if (string_compare (s, file_cmd_strings::COUNTRYSTR, 0, 8) == 0)
    options.country = convert_country(s.substr(8));
  else if (string_compare (s, file_cmd_strings::YEARSTR, 0, 5) == 0)
    options.year = convert_num(s, 5);
  else if (string_compare (s, file_cmd_strings::TYPESTR, 0, 5) == 0)
    options.type = maketype (s.substr(5));
  else if (string_compare (s, file_cmd_strings::AXLESSTR, 0, 6) == 0)
    options.axles = convert_num(s, 6);
  else if (string_compare (s, file_cmd_strings::DEBUGSTR, 0, 6) == 0)
    options.debug = convert_num(s, 6);
  else if (string_compare (s, file_cmd_strings::HISTORYSTR, 0, 8) == 0)
    options.history_file = s.substr(8);
  else return false;
  return true;
}


void
read_file (
  const char* filename,
  optionsT& options,
  filetypeT tp
)
{
  if (filename == NULL) throw ill_value (ill_value::INTERNAL);
	ifstream f (filename);
	if (!f) throw ill_file(filename);
  print_debug (cerr, 0, options, "reading file: ", filename);
	unsigned line = 0;
  string s;
	s.reserve (BUFSIZE);
	while (!f.eof () && f.good()) {
    try {
      getline(f, s);
      print_debug (cerr, 2, options, "reading line: ", s);
      trim_spaces(s);
		  ++line;
  		if (is_comment (s[0]) || s.empty()) continue; // comments	and empty lines
      else if (!read_commands(s, options, tp)) {
      	switch (tp) {
	    		case UIC_CAPITAL:
		    		 parse_capital (s, options);
			    	break;
  			  case UIC_SMALL:
	  			   parse_small (s, line, options);
  		  		break;	
	  		  case UIC_COUNTRY :
  	  			 parse_country (s, options);
	  	    	break;
          case UIC_REGIME :
            parse_regime (s, options);
            break;
          case UIC_SPEEDHEAT:
            parse_speed_heat(s, options);
            break;
          case UIC_MAINTYPE:
            parse_main_type(s, options);
            break;
          case UIC_IGA:
            parse_igad(s, line, options);
            break;
          default:
          case UIC_UNDEFINED :
            throw ill_value(ill_value::UNKNOWN);
          break;
         }
  		}
    }
    catch (file_error& err) {
      db_line_error(filename, err, line);      // XXX
    }
    catch (...) {
      cerr << "Unhandled exeption: " << line << endl;
    }
	}
}

