/*******************************************************************************
    riv2ascii
    Copyright (C) 2003  Harm Schotanus

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*******************************************************************************/


#include <iostream>
#include <string>
#include <fstream>
#include <cctype>
#include <sstream>
#include <cstdio>
#include <readline/readline.h>
#include <readline/history.h>
#include "riv2ascii.h"
#include "r2a_helper.h"
#include "r2a_prog_error.h"
#include "r2a_country.h"
#include "r2a_sub.h"
#include "r2a_main.h"
#include "r2a_uicnr.h"
#include "r2a_axles.h"
#include "riv_error.h"
#include "r2a_files.h"
#include "r2a_cmdln.h"
#include "r2a_regime.h"
#include "r2a_country_list.h"
#include "r2a_sub_list.h"
#include "r2a_main_list.h"
#include "r2a_regime_list.h"
#include "r2a_speed_heat_list.h"
#include "r2a_main_type_list.h"
#include "r2a_iga.h"
#include "r2a_iga_list.h"
#include "r2a_html.h"
#include "r2a_handle.h"
#include "r2a_file_error.h"
#include "r2a_notfound_error.h"


main_class_list main_classes;
country_code_list country_codes;
regime_list regimes;
speed_heat_list speeds_heats;
main_type_list main_types;
iga_list igas;


using std::ostringstream;
using std::cin;
using std::istream;
using std::ifstream;
using std::cerr;
using std::cout;


namespace {


void
match_string(
  ostream& out,
  const string& s,
  const unsigned axles=0,
  const unsigned year=0,
	const unsigned cntry=UIC,
  const typeT tp=unknown,
  const bool html=false
)
{
	put_table p(out, html);
  main_class_list ml;
	string::const_iterator si = s.begin();

  main_classes.find_match_c(s, *si, axles, year, ml);
  ml.write(out, *si, axles, year, html);
  ++si;
  sub_class_list sl;
  string::const_iterator se = s.end();
  for (; si != se ; ++si)	{
	  char l = *si;
 		unsigned doubled = 0;
  	for (;(si + 1) != se && *(si + 1) == l; ++si) ++doubled;
    if (*si == '-') continue;
    handleT <main_class> m (ml.first());
    m->find_match_s(sl, s, *si, doubled, cntry, year, axles, tp);
  }
  sl.write(out, axles, year, cntry, tp, html);
}


void
print_database(
  ostream& out,
  const bool html=false,
  const char c=0
)
{
  put_head p (out, html, "RIV2ASCII Database");
  switch (c) {
  case 'C':
  	main_classes.print(out, html);
    break;
  case 'c':
  	country_codes.print(out, html);
    break;
  case 'r':
    regimes.print(out, html);
    break;
  case 's':
    speeds_heats.print (out, html);
    break;
  case 'm':
    main_types.print(out, html);
    break;
  case 'i':
    igas.print(out, html);
    break;
  default:
    main_classes.print(out, html);
    country_codes.print(out, html);
    regimes.print(out, html);
    speeds_heats.print (out, html);
    main_types.print(out, html);
    igas.print(out, html);
    break;
  }
}


int
match_str(
  ostream& out,
  const string& s,
  optionsT& options
)
{
  try {
    string ostr;
    ostr.reserve(BUFSIZE);
    ostringstream outstr(ostr);
	  match_string (outstr, s, options.axles, options.year, options.country, options.type, options.html);
    out << outstr.str();
  }
	catch (inv_letter& l){
    cerr << l;
  }
  catch (letter_not_found& l) {
	  cerr << l;
  }  
  catch (riv_error& l){
    l.print(cerr);
  }
  return 0;
}


inline void
clear_database (
  const char  c = 0
)
{
  switch (c) {
  case 'h' :
    clear_history();
    break;
  case 'c' :
    country_codes.clear();
    break;
  case 'C' :
    main_classes.clear();
    break;
  case 'r' :
    regimes.clear();
    break;
  case 's':
    speeds_heats.clear();
    break;
  case 'm':
    main_types.clear();
    break;
  case 'i':
    igas.clear();
  default:
    country_codes.clear();
    regimes.clear();
    main_types.clear();
    speeds_heats.clear();
    main_classes.clear();
    igas.clear();
    break;
  }
}


int
do_iga(
  ostream& out,
  const string& s,
  optionsT& options
)
{
  iga_list il;
  try {
    igas.find(s, options.country, options.year, il);
    il.print(out, options.html);
  }
  catch (riv_error& l) {
    l.print(cerr);
  }
  return 0;
}


inline void
do_show(
  ostream& out,
  const optionsT& options
)
{
  out << "Axles: " << options.axles << endl
      << "Year: "<< options.year << endl
      << "Country: " << options.country << endl
      << "Type: " << type2str(options.type) << endl
      << "HTML: " << (options.html?"Yes":"No") << endl
      << "Show IGA: " << (options.show_iga?"Yes":"No") << endl
      << "Mode: " << char(options.default_mode) << endl
      << "Debug level: " << options.debug << endl
      << "History file: " << options.history_file << endl;
}


inline void
do_country(
  const string& s,
  optionsT& options
)
{
	if (isdigit (s[0]))
    options.country = convert_num (s, 0);
    if (options.country >= 100) options.country = 0;
	else {
    try
    {
      options.country = country_codes.find_country (s);
    }
    catch (country_not_found& l)
    {
      options.country = 0;
      cerr << l;
    }
  }
}


inline void
do_load(
  const string& s,
  optionsT& options
)
{
  try
  {
    read_file(s.c_str(), options);
  }
  catch (ill_file& fn)
  {
    cerr << fn;
  }
}


void
match_from_file (
  ostream& out,
  optionsT& options,
  const string& filename
);


inline void
do_read(
  ostream& out,
  optionsT& options,
  const string& s
)
{
  match_from_file (out, options, s);
}


namespace cmd_strings {


const string AXLESSTR ("axles=");
const string CLEARSTR ("clear=");
const string CLEARALLSTR ("clearall");
const string HELPSTR ("help");
const string YEARSTR ("year=");
const string TYPESTR ("type=");
const string HTMLSTR ("html=");
const string VERSIONSTR ("version");
const string PRINTALLSTR ("print");
const string PRINTSTR ("print=");
const string LOADSTR ("load=");
const string COUNTRYSTR ("country=");
const string SHOWSTR ("show");
const string QUITSTR ("quit");
const string READSTR ("read=");
const string DOIGASTR ("doiga=");
const string MODESTR ("mode=");
const string DEBUGSTR ("debug=");
const string HISTORYSTR ("history=");


} // namespace


modeT
make_mode (const char l)
{
  switch (l){
  case 'c' : return COMMAND;
  case 'r' : return MATCH;
  case 'i' : return IGA;
  case 'w' : return AXLES;
  case 'u' : return NUMBER;
  default :
    print_unknown_mode (cerr, l);
    return NONE;
  }
}


int
do_command (
  ostream& out,
  const string& s,
  optionsT& options
)
{
  if (!string_compare (s, cmd_strings::AXLESSTR, 0, 6)) options.axles = convert_num(s, 76);
  else if (!string_compare (s, cmd_strings::CLEARALLSTR, 0, 8)) clear_database();
  else if (!string_compare (s, cmd_strings::CLEARSTR, 0, 6)) clear_database(s[6]);
  else if (!string_compare (s, cmd_strings::HELPSTR, 0, 4)) give_help();
	else if (!string_compare (s, cmd_strings::YEARSTR, 0, 5)) options.year = convert_num(s, 5);
	else if (!string_compare (s, cmd_strings::TYPESTR, 0, 5)) options.type = maketype(s[5]);
	else if (!string_compare (s, cmd_strings::HTMLSTR, 0, 5)) options.html = parse_bool(s, 5);
	else if (!string_compare (s, cmd_strings::VERSIONSTR, 0, 7)) display_version();
  else if (!string_compare (s, cmd_strings::PRINTSTR, 0, 6)) print_database(out, options.html, s[6]);
  else if (!string_compare (s, cmd_strings::PRINTALLSTR, 0, 5)) print_database(out, options.html);
  else if (!string_compare (s, cmd_strings::LOADSTR, 0, 5))  do_load(s.substr( 5), options);
  else if (!string_compare (s, cmd_strings::COUNTRYSTR, 0, 8)) do_country(s.substr(8), options);
 	else if (!string_compare (s, cmd_strings::QUITSTR, 0, 4)) return 1;
	else if (!string_compare (s, cmd_strings::SHOWSTR, 0, 4)) do_show(out, options);
  else if (!string_compare (s, cmd_strings::READSTR, 0, 5)) do_read(out, options, s.substr(5));
  else if (!string_compare (s, cmd_strings::HISTORYSTR, 0, 8)) options.history_file = s.substr(8);
  else if (!string_compare (s, cmd_strings::DOIGASTR, 0, 6)) options.show_iga = parse_bool(s, 6);
  else if (!string_compare (s, cmd_strings::DEBUGSTR, 0, 6)) options.debug = convert_num(s, 6);
  else if (!string_compare (s, cmd_strings::MODESTR, 0, 5)) options.default_mode = make_mode (s[5]);
  else print_unknown_command(cerr, s);
  return 0;
}


int
do_uicnumber(
  ostream& out,
  const string& s,
  optionsT& options
)
{
  try {
    uic_number* u = parse_code(s);
    u->print_number(out, options.html);
    delete u;
  }
  catch (inv_uic_nr& l)
  {
    cerr << l;
  }
  catch (not_found& l)
  {
    cerr << l;
  }
  catch (riv_error& l)
  {
    l.print(cerr);
  }
  catch (...) {}
  return 0;
}


int
do_axles(
  ostream& out,
  const string& s,
  optionsT& options
)
{
  try {
    axle_set axle (s);
    axle.print(out, options.html);
  }
  catch (riv_error& l) {
    l.print(cerr);
  }
  return 0;
}


struct {
  modeT m;
  int (*f)(ostream&, const string&, optionsT&) ;
} f_table []=
{
  {MATCH, match_str},
  {COMMAND, do_command},
  {AXLES, do_axles},
  {NUMBER, do_uicnumber},
  {IGA, do_iga},
  {NONE, do_iga}
};


int do_f_table(
  ostream& out,
  const string& s,
  modeT m,
  optionsT& options
)
{
  for (int i = 0; f_table[i].m != NONE; ++i)
    if (m == f_table[i].m) return f_table[i].f(out, s, options);
  throw ill_value(ill_value::MODE);
}


int read_line(
  ostream& out,
  const char letter,
  string& s,
  optionsT& options
)
{
  if (letter < 32) throw not_found(to_string(letter));
  trim_spaces (s);
	switch (letter) {
  case '#' :
  case '%' :
  case '@' :
  case ':' :
  case '$' :
    return do_f_table (out, s.substr(1), static_cast<modeT> (letter), options);
    break;
  case '?' :
    give_help();
    break;
  default:
    do_f_table(out, s, options.default_mode, options);
    break;
  }
  return 0;
}


void
interactive (
  ostream& out,
  optionsT& options
)
{
  using_history();
  stifle_history(HISTORY_SIZE);
  if (!options.history_file.empty()) read_history(options.history_file.c_str());
  int l = 0;
	do {
    char *t = readline(prompt);
    if (t != 0 && *t != 0) {
      add_history(t);
      string s(t);
      try {
        l = read_line(out, *t, s, options);
      }
      catch (riv_error& err) {
        print_ill_mode (cerr, s);
      }
			catch (...) { cerr << "Uncaught exception\n";}
    }
    free(t);

	}	while (!l);	
  if (!options.history_file.empty()) write_history(options.history_file.c_str());				
}


void
match_from_file (
  ostream& out,
  optionsT& options,
  const string& filename
)
{
  ifstream f (filename.c_str());
	if (!f) throw ill_file(filename);
	unsigned line = 0;
  string s;
	s.reserve (LINE_SIZE);
	while (!f.eof () && f.good()) {
    try {
      ++line;
      getline (f, s);
      if (read_line(out, s[0], s, options)) break;
      out << endl;
    }
    catch (riv_error& err) {
      db_line_error (options.file, err, line);
    }
  }
}


}// namespace


int
main (
  const int argc,
  char** argv
)
{
	optionsT options;
  try {
    read_file("riv2ascii.rc", options);
  }
  catch (ill_file fn){
    cerr << fn;
  }
 	parse_options (argc, argv, options);
  try {
    if (!options.s_file) read_file ("files.rc", options);
	  if (options.find_cnt)
		  options.country = country_codes.find_country (options.company_name);
  }
  catch (ill_file& fn){
    cerr << fn;
  }
  catch (not_found& fn){
    cerr << fn;
  }

  switch  (options.mode) {
    default:
    break;
  case NUMBER:
    do_uicnumber(cout, options.uic_code, options);
    break;
  case AXLES:
    do_axles(cout, options.axles_arr, options);
    break;
  case PRINT:
    print_database(cout, options.html);
    break;
	case MATCH:
    {
		  put_head p(cout, options.html, "RIV2ASCII :", options.riv_text);
		  match_str (cout, options.riv_text, options);
    }
		break;
	case INTERACTIVE:
		interactive(cout, options);
 		break;
  case FILE_MODE:
    match_from_file (cout, options, options.file);
    break;
	case CODE:
		break;
  case IGA:
    {
      put_head p(cout, options.html, "RIV2ASCII :", options.riv_text);
      do_iga(cout, options.riv_text, options);
    }
    break;
	}
 	cout.flush();
  cerr.flush();
	return 0;
}

